from cmakelang.parse.additional_nodes import ShellCommandNode
from cmakelang.parse.argument_nodes import (
    PositionalParser, StandardArgTree, StandardParser)


def parse_external_project_add(ctx, tokens, breakstack):
  """
  ::

    ExternalProject_Add(<name> [<option>...])

  :see: https://cmake.org/cmake/help/v3.14/module/ExternalProject.html
  """
  return StandardArgTree.parse(
      ctx, tokens,
      npargs=1,
      kwargs={
          # Directory Options
          "PREFIX": PositionalParser(1),
          "TMP_DIR": PositionalParser(1),
          "STAMP_DIR": PositionalParser(1),
          "LOG_DIR": PositionalParser(1),
          "DOWNLOAD_DIR": PositionalParser(1),
          "SOURCE_DIR": PositionalParser(1),
          "BINARY_DIR": PositionalParser(1),
          "INSTALL_DIR": PositionalParser(1),
          # Download Step Options
          "DOWNLOAD_COMMAND": ShellCommandNode.parse,
          "URL": PositionalParser('+'),
          "URL_HASH": PositionalParser(1),
          "URL_MD5": PositionalParser(1),
          "DOWNLOAD_NAME": PositionalParser(1),
          "DOWNLOAD_NO_EXTRACT": PositionalParser(1),
          "DOWNLOAD_NO_PROGRESS": PositionalParser(1),
          "TIMEOUT": PositionalParser(1),
          "HTTP_USERNAME": PositionalParser(1),
          "HTTP_PASSWORD": PositionalParser(1),
          "HTTP_HEADER": PositionalParser('+'),
          "TLS_VERIFY": PositionalParser(1),
          "TLS_CAINFO": PositionalParser(1),
          "NETRC": PositionalParser(
              1, flags=["CMAKE_NETRC", "IGNORED", "OPTIONAL", "REQUIRED"]),
          "NETRC_FILE": PositionalParser(1),
          # Git
          "GIT_REPOSITORY": PositionalParser(1),
          "GIT_TAG": PositionalParser(1),
          "GIT_REMOTE_NAME": PositionalParser(1),
          "GIT_SUBMODULES": PositionalParser('+'),
          "GIT_SHALLOW": PositionalParser(1),
          "GIT_PROGRESS": PositionalParser(1),
          "GIT_CONFIG": PositionalParser('+'),
          # Subversion
          "SVN_REPOSITORY": PositionalParser(1),
          "SVN_REVISION": PositionalParser(1),
          "SVN_USERNAME": PositionalParser(1),
          "SVN_PASSWORD": PositionalParser(1),
          "SVN_TRUST_CERT": PositionalParser(1),
          # Mercurial
          "HG_REPOSITORY": PositionalParser(1),
          "HG_TAG": PositionalParser(1),
          # CVS
          "CVS_REPOSITORY": PositionalParser(1),
          "CVS_MODULE": PositionalParser(1),
          "CVS_TAG": PositionalParser(1),
          # Update/Patch Step Options
          "UPDATE_COMMAND": ShellCommandNode.parse,
          "UPDATE_DISCONNECTED": PositionalParser(1),
          "PATCH_COMMAND": ShellCommandNode.parse,
          # Configure Step Options
          "CONFIGURE_COMMAND": ShellCommandNode.parse,
          "CMAKE_COMMAND": PositionalParser(1),
          "CMAKE_GENERATOR": PositionalParser(1),
          "CMAKE_GENERATOR_PLATFORM": PositionalParser(1),
          "CMAKE_GENERATOR_TOOLSET": PositionalParser(1),
          "CMAKE_GENERATOR_INSTANCE": PositionalParser(1),
          "CMAKE_ARGS": PositionalParser('+'),
          "CMAKE_CACHE_ARGS": PositionalParser('+'),
          "CMAKE_CACHE_DEFAULT_ARGS": PositionalParser('+'),
          "SOURCE_SUBDIR": PositionalParser(1),
          # Build Step Options
          "BUILD_COMMAND": ShellCommandNode.parse,
          "BUILD_IN_SOURCE": PositionalParser(1),
          "BUILD_ALWAYS": PositionalParser(1),
          "BUILD_BYPRODUCTS": PositionalParser('+'),
          # Install Step Options
          "INSTALL_COMMAND": ShellCommandNode.parse,
          # Test Step Options
          "TEST_COMMAND": ShellCommandNode.parse,
          "TEST_BEFORE_INSTALL": PositionalParser(1),
          "TEST_AFTER_INSTALL": PositionalParser(1),
          "TEST_EXCLUDE_FROM_MAIN": PositionalParser(1),
          # Output Logging Options
          "LOG_DOWNLOAD": PositionalParser(1),
          "LOG_UPDATE": PositionalParser(1),
          "LOG_PATCH": PositionalParser(1),
          "LOG_CONFIGURE": PositionalParser(1),
          "LOG_BUILD": PositionalParser(1),
          "LOG_INSTALL": PositionalParser(1),
          "LOG_TEST": PositionalParser(1),
          "LOG_MERGED_STDOUTERR": PositionalParser(1),
          "LOG_OUTPUT_ON_FAILURE": PositionalParser(1),
          # Terminal Access Options
          "USES_TERMINAL_DOWNLOAD": PositionalParser(1),
          "USES_TERMINAL_UPDATE": PositionalParser(1),
          "USES_TERMINAL_CONFIGURE": PositionalParser(1),
          "USES_TERMINAL_BUILD": PositionalParser(1),
          "USES_TERMINAL_INSTALL": PositionalParser(1),
          "USES_TERMINAL_TEST": PositionalParser(1),
          # Target Options
          "DEPENDS": PositionalParser('+'),
          "EXCLUDE_FROM_ALL": PositionalParser(1),
          "STEP_TARGETS": PositionalParser('+'),
          "INDEPENDENT_STEP_TARGETS": PositionalParser('+'),
          # Misc Options
          "LIST_SEPARATOR": PositionalParser(1),
          "COMMAND": ShellCommandNode.parse,
      },
      flags=[],
      breakstack=breakstack)


def parse_external_project_add_step(ctx, tokens, breakstack):
  """
  ::

    ExternalProject_Add_Step(<name> [<option>...])

  :see: https://cmake.org/cmake/help/v3.14/module/ExternalProject.html#command:externalproject_add_step
  """
  return StandardArgTree.parse(
      ctx, tokens,
      npargs=2,
      kwargs={
          "COMMAND": ShellCommandNode.parse,
          "COMMENT": PositionalParser('+'),
          "DEPENDEES": PositionalParser('+'),
          "DEPENDERS": PositionalParser('+'),
          "DEPENDS": PositionalParser('+'),
          "BYPRODUCTS": PositionalParser('+'),
          "ALWAYS": PositionalParser(1),
          "EXCLUDE_FROM_MAIN": PositionalParser(1),
          "WORKING_DIRECTORY": PositionalParser(1),
          "LOG": PositionalParser(1),
          "USES_TERMINAL": PositionalParser(1),
      },
      flags=[],
      breakstack=breakstack)


def populate_db(parse_db):
  parse_db["externalproject_add"] = parse_external_project_add
  parse_db["externalproject_get_property"] = StandardParser('+')
  parse_db["externalproject_add_step"] = parse_external_project_add_step
  parse_db["externalproject_add_steptargets"] = \
      StandardParser('+', flags=["NO_DEPENDS"])
  parse_db["externalproject_add_stepdependencies"] = StandardParser('+')



# -*- coding: utf-8 -*-
# pylint: disable=too-many-lines
from __future__ import print_function
from __future__ import unicode_literals

import logging

from cmakelang import common
from cmakelang import lex
from cmakelang.parse.util import (
    is_semantic_token, is_syntactic_token, is_whitespace_token, is_comment_token
)

logger = logging.getLogger(__name__)


class NodeType(common.EnumObject):
  """
  Enumeration for AST nodes
  """
  _id_map = {}


NodeType.BODY = NodeType(0)
NodeType.WHITESPACE = NodeType(1)
NodeType.COMMENT = NodeType(2)
NodeType.STATEMENT = NodeType(3)
NodeType.FLOW_CONTROL = NodeType(4)
NodeType.FUNNAME = NodeType(10)
NodeType.ARGGROUP = NodeType(5)
NodeType.KWARGGROUP = NodeType(6)
NodeType.PARGGROUP = NodeType(14)
NodeType.FLAGGROUP = NodeType(15)
NodeType.PARENGROUP = NodeType(16)
NodeType.ARGUMENT = NodeType(7)
NodeType.KEYWORD = NodeType(8)
NodeType.FLAG = NodeType(9)
NodeType.ONOFFSWITCH = NodeType(11)
NodeType.ATWORDSTATEMENT = NodeType(17)
NodeType.ATWORD = NodeType(18)

# NOTE(josh): These aren't really semantic, but they have structural
# significance that is important in formatting. Since they will have a presence
# in the format tree, we give them a presence in the parse tree as well.
NodeType.LPAREN = NodeType(12)
NodeType.RPAREN = NodeType(13)


class FlowType(common.EnumObject):
  """
  Enumeration for flow control types
  """
  _id_map = {}


FlowType.IF = FlowType(0)
FlowType.WHILE = FlowType(1)
FlowType.FOREACH = FlowType(2)
FlowType.FUNCTION = FlowType(3)
FlowType.MACRO = FlowType(4)


class TreeNode(object):
  """
  A node in the full-syntax-tree.
  """

  def __init__(self, node_type):
    self.node_type = node_type
    self.children = []
    self.parent = None

  def build_ancestry(self):
    """Recursively assign the .parent member within the subtree."""
    for child in self.children:
      if isinstance(child, TreeNode):
        child.parent = self
        child.build_ancestry()

  def get_location(self):
    """
    Return the (line, col) of the first token in the subtree rooted at this
    node.
    """
    if self.children:
      return self.children[0].get_location()

    return lex.SourceLocation((0, 0, 0))

  def count_newlines(self):
    newline_count = 0
    for child in self.children:
      newline_count += child.count_newlines()
    return newline_count

  def __repr__(self):
    return '{}: {}'.format(self.__class__.__name__, self.get_location())

  def get_tokens(self, out=None, kind=None):
    if out is None:
      out = []
    if kind is None:
      kind = "all"

    match_group = {
        "semantic": is_semantic_token,
        "syntactic": is_syntactic_token,
        "whitespace": is_whitespace_token,
        "comment": is_comment_token,
        "all": lambda x: True
    }[kind]

    for child in self.children:
      if isinstance(child, lex.Token):
        if match_group(child):
          out.append(child)
      elif isinstance(child, TreeNode):
        child.get_tokens(out, kind)
      else:
        raise RuntimeError("Unexpected child of type {}".format(type(child)))
    return out

  def get_semantic_tokens(self, out=None):
    """
    Recursively reconstruct a stream of semantic tokens
    """
    return self.get_tokens(out, kind="semantic")


class ParenBreaker(object):
  """
  Callable that returns true if the supplied token is a right parenthential
  """

  def __call__(self, token):
    return token.type == lex.TokenType.RIGHT_PAREN


class KwargBreaker(object):
  """
  Callable that returns true if the supplied token is in the list of keywords,
  ignoring case.
  """

  def __init__(self, kwargs):
    self.kwargs = [kwarg.upper() for kwarg in kwargs]

  def __call__(self, token):
    return token.spelling.upper() in self.kwargs

  def __repr__(self):
    return "KwargBreaker({})".format(",".join(self.kwargs))



with section("lint"):
  disabled_codes = [
    # A custom command with one output doesn't really need a comment because
    # the default "generating XXX" is a good message already.
    "C0113",
  ]


with section("parse"):
  additional_commands = {
    # Wrappers.cmake
    "cc_binary": {
      "pargs": "1+",
      "flags": ["ADD_RUNTARGET"],
      "kwargs": {
          "SRCS": "*",
          "DEPS": "*",
          "PKGDEPS": "*",
          "PROPERTIES": {
            "kwargs": {
              "OUTPUT_NAME": 1,
              "EXPORT_NAME": 1,
            }
          }
      }
    },

    "cc_library": {
      "pargs": "1+",
      "flags": ["STATIC", "SHARED"],
      "kwargs": {
          "INC": {
            "pargs": 0,
            "kwargs": {
                "PUBLIC": "*",
                "PRIVATE": "*",
                "INTERFACE": "*",
            }
          },
          "LIBDIRS": {
            "pargs": "*",
            "kwargs": {
                "PUBLIC": "*",
                "PRIVATE": "*",
                "INTERFACE": "*",
            }
          },
          "SRCS": "*",
          "DEPS": {
            "pargs": "*",
            "kwargs": {
                "PUBLIC": "*",
                "PRIVATE": "*",
                "INTERFACE": "*",
            }
          },
          "PKGDEPS": "*",
          "PROPERTIES": {
            "kwargs": {
              "ARCHIVE_OUTPUT_NAME": 1,
              "EXPORT_NAME": 1,
              "INTERFACE_INCLUDE_DIRECTORIES": 1,
              "LIBRARY_OUTPUT_NAME": 1,
              "OUTPUT_NAME": 1,
              "SOVERSION": 1,
              "SUFFIX": 1,
              "VERSION": 1,
            }
          }
      }
    },

    "cc_test": {
      "pargs": 1,
      "kwargs": {
          "SRCS": "*",
          "DEPS": "*",
          "PKGDEPS": "*",
          "ARGV": "*",
          "TEST_DEPS": "*",
          "WORKING_DIRECTORY": "*",
          "LABELS": "*"
      }
    },

    "check_call": {
      "kwargs": {
        "COMMAND": "*",
        "WORKING_DIRECTORY": "1",
        "TIMEOUT": "1",
        "RESULT_VARIABLE": "1",
        "RESULTS_VARIABLE": "1",
        "OUTPUT_VARIABLE": "1",
        "ERROR_VARIABLE": "1",
        "INPUT_FILE": "1",
        "OUTPUT_FILE": "1",
        "ERROR_FILE": "1",
        "ENCODING": "1",
      },
      "flags": [
        "OUTPUT_QUIET",
        "ERROR_QUIET",
        "OUTPUT_STRIP_TRAILING_WHITESPACE",
        "ERROR_STRIP_TRAILING_WHITESPACE",
      ]
    },

    "join": {
      "pargs": [1, "+"],
      "kwargs": {
          "GLUE": 1,
      }
    },

    "check_pyoneline": {
      "pargs": "+",
      "kwargs": {
        "OUTPUT_VARIABLE": 1,
        "ERROR_VARIABLE": 1
      }
    },

    "tangent_addtest": {
      "kwargs": {
        "NAME": 1,
        "WORKING_DIRECTORY": 1,
        "COMMAND": "+",
        "CONFIGURATIONS": "+"
      }
    },

    "travis_decrypt": {
      "pargs": [3],
      "kwargs": {}
    },

    # debian.cmake
    "create_debian_binary_packages": {
      "pargs": [3, "+"],
      "kwargs": {
        "OUTPUTS": "*",
        "DEPS": "*"
      }
    },
    "create_debian_packages": {
      "pargs": [
        {"nargs": "+", "flags": ["FORCE_PBUILDER"]},
      ],
      "kwargs": {
        "OUTPUTS": "*",
        "DEPS": "*"
      }
    },
    "debhelp": {
      "pargs": ["1+", ],
      "spelling": "DEBHELP"
    },
    "get_debs": {
      "pargs": [3, "*"],
    },

    # pkgconfig.cmake
    "pkg_find": {
      "kwargs": {
        "PKG": "*"
      }
    },

    # codestyle.cmake
    "format_and_lint": {
      "kwargs": {
        "CC": "*",
        "CMAKE": "*",
        "PY": "*",
        "JS": "*",
        "SHELL": "*"
      }
    },

    # other:
    "importvars": {
      "pargs": "1+",
      "kwargs": {
        "VARS": "+"
      },
      "spelling": "IMPORTVARS"
    },

    "exportvars": {
      "pargs": "1+",
      "kwargs": {
        "VARS": "+"
      },
      "spelling": "EXPORTVARS"
    },

    "stage_files": {
      "kwargs": {
        "LIST": 1,
        "STAGE": 1,
        "SOURCEDIR": 1,
        "FILES": "*"
      },
    },


    # GtkDocConfig.cmake
    "gtk_doc_add_module": {
      "pargs": 1,
      "kwargs": {
        "FIXREFOPTS": "*",
        "IGNOREHEADERS": "*",
        "LIBRARIES": "*",
        "LIBRARY_DIRS": "*",
        "SOURCE": "*",
        "SUFFIXES": "*",
        "XML": 1,
      },
    },

  }



"""
Higher level commands for github pipeline
"""

from __future__ import absolute_import, unicode_literals

import argparse
import io
import logging
import os
import pipes
import random
import shutil
import subprocess
import sys
import tempfile

if True:
  if os.path.samefile(sys.path[0], os.path.dirname(__file__)):
    sys.path.pop(0)

  import github
  import keyring
  from packaging import version as versiontool
  import magic

logger = logging.getLogger(__name__)

PSEUDO_MESSAGE = """
This is a pseudo-release used only to stage artifacts for the release pipeline.
Please do not rely on the artifacts contained in this release as they change
frequently and are very likely to be broken.
""".replace("\n", " ")


def get_access_token():
  if "GITHUB_ACCESS_TOKEN" in os.environ:
    return os.environ["GITHUB_ACCESS_TOKEN"]

  credential = keyring.get_credential("api.github.com", None)
  if credential is not None:
    return credential.password

  raise RuntimeError(
      "No github access token found, please set GITHUB_ACCESS_TOKEN "
      "environment variable or add username@api.github.com to your "
      "system keyring, e.g. run `keyring set api.github.com <username>`")


def create_pseudorelease_tag(reposlug, branch):
  # TODO(josh): get title out of the script so that it can be reusable
  hub = github.Github(get_access_token())
  repo = hub.get_repo(reposlug)
  branchobj = repo.get_branch(branch)
  logger.info("Creating tag pseudo-%s -> %s", branch, branchobj.commit.sha)
  refname = "tags/pseudo-" + branch
  try:
    existing_ref = repo.get_git_ref(refname)
    logger.info("Updating existing ref for %s", refname)
    existing_ref.edit(branchobj.commit.sha, force=True)
    return
  except github.UnknownObjectException:
    pass

  logger.info("Creating ref %s", refname)
  refname = "refs/" + refname
  repo.create_git_ref(refname, branchobj.commit.sha)


INITDIRS = [
    "branches",
    "hooks",
    "info",
    "objects/info",
    "objects/pack",
    "refs/heads",
    "refs/tags",
]

CONFIGTPL = """
[core]
	repositoryformatversion = 0
	filemode = true
	bare = true
[remote "origin"]
	url = {}
  fetch = +refs/heads/*:refs/heads/*
"""

MSGTPL = """Sync {commit_name}

This commit was automatically generated by the upstream build.

Upstream-Commit: {upstream_commit}
"""

CHARS = "abcdefghijklmnopqrstuvwxzyABCDEFGHIJKLMNOPQRSTUVWXYZ012345679"


def make_randstr(strlen=10):
  """Generate a random string."""
  return ''.join(random.choice(CHARS) for i in range(strlen))


def init_repository(repodir, repo_url):
  """
  Ensure that the specified directory is configured as a git repository.
  We create any missing directories and overwrite the config and HEAD files,
  putting the repository directory into a known state without blowing away
  the object cache.
  """

  # Ensure that the standard directories exit
  for dirname in INITDIRS:
    dirpath = os.path.join(repodir, dirname)
    if os.path.lexists(dirpath):
      if os.path.islink(dirpath) or not os.path.isdir(dirpath):
        os.unlink(dirpath)
      else:
        continue
    os.makedirs(dirpath)

  # Write out the config and the HEAD files
  outpath = os.path.join(repodir, "config")
  with io.open(outpath, "w", encoding="utf-8") as outfile:
    outfile.write(CONFIGTPL.format(repo_url))
  outpath = os.path.join(repodir, "HEAD")
  with io.open(outpath, "w", encoding="utf-8") as outfile:
    outfile.write("ref: refs/heads/master")


def canonicalize_trailer_key(keystr):
  """Return canonical capitalization for a trailer key"""
  return "-".join(part.capitalize() for part in keystr.split("-"))


def get_trailers(repodir, commit):
  """Return a dictionary of git commit trailers."""
  proc = subprocess.Popen([
      "git", "show", "--no-patch", '--format="%(trailers:unfold)',
      commit], cwd=repodir, stdout=subprocess.PIPE)

  out = {}
  with proc.stdout as stdout:
    for line in stdout:
      line = line.strip().decode("utf-8")
      if ":" not in line:
        continue
      key, value = line.split(":", 1)
      if ", " in value:
        value = value.split(", ")
      out[canonicalize_trailer_key(key)] = value
  proc.wait()
  if proc.returncode != 0:
    cmdstr = " ".join(pipes.quote(arg) for arg in proc.args)
    raise subprocess.CalledProcessError(proc.returncode, cmdstr)
  return out


def ref_exists(repodir, refname):
  return subprocess.call(
      ["git", "show-ref", "--verify", "--quiet", "refs/" + refname],
      cwd=repodir) == 0


def sync_doc_artifacts(
    docrepo_url, docrepo_dir, scratch_dir, stage_dir, deploy_key_path,
    branch=None, tag=None):

  current_commit = subprocess.check_output(
      ["git", "rev-parse", "HEAD"]).strip().decode("utf-8")

  env = os.environ.copy()
  if deploy_key_path:
    env["GIT_SSH_COMMAND"] = "ssh -i {} -F /dev/null".format(deploy_key_path)

  # Initialize and fetch the docrepo clone
  init_repository(docrepo_dir, docrepo_url)
  subprocess.check_call([
      "git", "fetch", "--prune", "--tags"], cwd=docrepo_dir, env=env)

  # Create a temporary branchname for us to do our work in
  workbranch = "work-" + make_randstr(10)

  # Clear out scratch tree
  if os.path.exists(scratch_dir):
    shutil.rmtree(scratch_dir)
  os.makedirs(scratch_dir)

  if branch:
    commit_name = branch
    if ref_exists(docrepo_dir, "heads/" + branch):
      # If the branch already exists on the remote, then check it out at
      # it's current state
      subprocess.check_call(
          ["git", "--git-dir=" + docrepo_dir, "--work-tree=" + scratch_dir,
           "checkout", "-b", workbranch, branch])
    else:
      # Otherwise create a new orphan branch for it
      subprocess.check_call(
          ["git", "--git-dir=" + docrepo_dir, "--work-tree=" + scratch_dir,
           "checkout", "--orphan", workbranch])
  elif tag:
    commit_name = tag
    if ref_exists(docrepo_dir, "tags/" + tag):
      # If the tag already exists on the remote then checkout it's parent
      subprocess.check_call(
          ["git", "--git-dir=" + docrepo_dir, "--work-tree=" + scratch_dir,
           "checkout", "-b", workbranch, "{}^".format(tag)])
    else:
      # Otherwise checkout the current HEAD of master and just use that.
      subprocess.check_call(
          ["git", "--git-dir=" + docrepo_dir, "--work-tree=" + scratch_dir,
           "checkout", "-b", workbranch, "master"])
  else:
    raise ValueError("Both branch and tag are empty")

  subprocess.check_call(
      ["git", "--git-dir=" + docrepo_dir, "--work-tree=" + stage_dir,
       "add", "-A"])

  result = subprocess.call(
      ["git", "diff", "--quiet", "--exit-code", "--cached"],
      cwd=docrepo_dir)
  if result == 0:
    # There is nothing to commit, the doc sources have not changed.
    return

  tfile = tempfile.NamedTemporaryFile(
      delete=False, prefix="gitmsg-", mode="w")
  with tfile as outfile:
    msgpath = outfile.name
    outfile.write(MSGTPL.format(
        commit_name=commit_name, upstream_commit=current_commit))
  subprocess.check_call([
      "git", "--git-dir=" + docrepo_dir, "--work-tree=" + stage_dir,
      "commit", "--file", msgpath])
  os.unlink(msgpath)

  if branch:
    subprocess.check_call(
        ["git", "push", "-f", "origin", "{}:{}".format(workbranch, branch)],
        cwd=docrepo_dir, env=env)
  elif tag:
    subprocess.check_call(
        ["git", "tag", "-f", tag],
        cwd=docrepo_dir)
    subprocess.check_call(
        ["git", "push", "-f", "origin", "{0}:{0}".format(tag)],
        cwd=docrepo_dir, env=env)


PSEUDO_MESSAGE = """\
This is a pseudo-release used only to stage artifacts for the release pipeline.
Please do not rely on the artifacts contained in this release as they change
frequently and are very likely to be broken.
""".replace("\n", " ")


def push_release(reposlug, tag, message_path, filepaths):
  message = ""
  if message_path:
    with io.open(message_path, encoding="utf-8") as infile:
      message = infile.read()

  # TODO(josh): get title out of the script so that it can be reusable
  prerelease = False
  title = reposlug.split("/", 1)[1].replace("_", "-") + " " + tag

  if tag.startswith("pseudo-"):
    prerelease = True
    title = "pseudo-release artifacts for " + tag[len("pseudo-"):]
  else:
    try:
      version = versiontool.parse(tag)
      if version.is_devrelease or version.is_prerelease:
        prerelease = True
    except versiontool.InvalidVersion:
      version = None

  filenames = set(filepath.rsplit(os.sep, 1)[-1] for filepath in filepaths)
  hub = github.Github(get_access_token())
  repo = hub.get_repo(reposlug)

  try:
    release = repo.get_release(tag)
    logger.info("Found release for tag %s", tag)
  except github.UnknownObjectException:
    logger.info("Creating release for tag %s", tag)
    try:
      release = repo.create_git_release(
          tag, title, message, prerelease=prerelease)
    except github.UnknownObjectException:
      logger.warning("Can't create a release for tag %s", tag)
      raise

  if release.title != title or release.body != message:
    logger.info("Updating release message for: %d", release.id)
    release.update_release(title, message, prerelease=prerelease)

  for asset in release.get_assets():
    if asset.name in filenames:
      logger.info("Deleting asset %s", asset.name)
      asset.delete_asset()

  for filepath in filepaths:
    filename = filepath.rsplit(os.sep, 1)[-1]
    mime = magic.Magic(mime=True)
    release.upload_asset(
        filepath, content_type=mime.from_file(filepath), name=filename)


def setup_argparser(argparser):
  argparser.add_argument(
      "-l", "--log-level", default="info",
      choices=["debug", "info", "warning", "error"])
  subparsers = argparser.add_subparsers(dest="command")

  subparser = subparsers.add_parser(
      "create-pseudorelease-tag", help=(
          "Create a tag pseudo-<branch> pointing to the current head of "
          "<branch> on github."
      ))
  subparser.add_argument("reposlug")
  subparser.add_argument("branch")

  subparser = subparsers.add_parser(
      "sync-doc-artifacts", help=(
          "Given the currently checked-out commit within an export repository,"
          " create a new commit in the corresponding documentation artifacts"
          " repository."
      ))
  subparser.add_argument(
      "--doc-repo", dest="docrepo_url",
      help="URL of the documentation repository")
  subparser.add_argument(
      "--repo-dir", dest="docrepo_dir",
      help="path to where we should checkout the doc artifacts repo")
  subparser.add_argument(
      "--scratch-tree", dest="scratch_dir",
      help="directory to use as git work-tree for intermediate steps")
  subparser.add_argument(
      "--stage", dest="stage_dir",
      help="directory that contains the new content for the repository")
  subparser.add_argument(
      "--deploy-key", dest="deploy_key_path", nargs="?",
      help="Path to the deploy key for the documentation repository")

  mgroup = subparser.add_mutually_exclusive_group(required=True)
  mgroup.add_argument(
      "--branch",
      help="Name of the target branch to push to")
  mgroup.add_argument(
      "--tag",
      help="Name of the target tag to push to")

  subparser = subparsers.add_parser(
      "push-release", help=(
          "Create or modify the release for the currently built tag on travis."
      ))

  subparser.add_argument(
      "-m", "--message",
      help="path to a file containing release notes message")
  subparser.add_argument("--repo-slug", required=True)
  subparser.add_argument("--tag", help="the tag we are deploying")
  subparser.add_argument("files", nargs="*", help="files to upload")


def get_argdict(namespace):
  blacklist = tuple()
  out = {}
  for key, value in vars(namespace).items():
    if key.startswith("_"):
      continue
    if key in blacklist:
      continue
    out[key] = value
  return out


def main():
  logging.basicConfig(level=logging.INFO)
  argparser = argparse.ArgumentParser(description=__doc__)
  setup_argparser(argparser)

  try:
    import argcomplete
    argcomplete.autocomplete(argparser)
  except ImportError:
    pass

  args = argparser.parse_args()
  logging.getLogger().setLevel(getattr(logging, args.log_level.upper()))

  if hasattr(args, "tag"):
    tagstr = getattr(args, "tag")
    travis_tag = os.environ.get("TRAVIS_TAG", tagstr)

    if travis_tag != tagstr:
      # If this is a travis incremental build, then stage artifacts under the
      # pseudo-release instead of creating a new release for the tag.
      if travis_tag.startswith("pseudo-"):
        setattr(args, "tag", travis_tag)
      else:
        logger.error(
            "Cowardly refusing to to create a github release '%s' from a travis"
            "build of tag '%s'", tagstr, os.environ["TRAVIS_TAG"])
        sys.exit(1)

  argdict = get_argdict(args)
  command = argdict.pop("command")
  argdict.pop("log_level")

  if command == "create-pseudorelease-tag":
    create_pseudorelease_tag(args.reposlug, args.branch)
  elif command == "sync-doc-artifacts":
    tag = argdict.pop("tag")
    if tag.startswith("pseudo-"):
      argdict["branch"] = tag[len("pseudo-"):]
    else:
      argdict["tag"] = tag
    sync_doc_artifacts(**argdict)
  elif command == "push-release":
    push_release(args.repo_slug, args.tag, args.message, args.files)
  else:
    logger.error("Unknown command %s", command)
    sys.exit(1)


if __name__ == "__main__":
  main()



# -*- coding: utf-8 -*-
"""
Helper scripts for building debian packages. Provides a number of frontend
commands that are called by the cmake generated build system to work through
the different stages of building the debian packages.
"""

from __future__ import print_function
from __future__ import unicode_literals

import argparse
import io
import logging
import math
import os
import re
import subprocess
import sys
import time

import enum
import requests

logger = logging.getLogger(__name__)


def parse_changelog(logpath):
  """
  Parse the changelog to get package name and version number
  """
  with io.open(logpath, "r", encoding="utf-8") as infile:
    lineiter = iter(infile)
    firstline = next(lineiter)

  pattern = r"(\S+) \(([^\)]+)\) (\S+); urgency=\S+"
  match = re.match(pattern, firstline)
  assert match, "Failed to match firstline pattern:\n {}".format(firstline)

  package_name = match.group(1)
  version = match.group(2)
  distribution = match.group(3)

  if "-" in version:
    upstream_version, debian_revision = version.rsplit("-", 1)
  else:
    upstream_version = version
    debian_revision = 0

  return package_name, upstream_version, debian_revision, distribution


class Flags(enum.IntEnum):
  EXCLUDE = 0b1
  DIRECTORY = 0b01


def should_include(patterns, relpath, query_flags=0):
  """
  Return true if a directory should be included
  """
  most_recent_positive_match = None

  output = False
  if query_flags & Flags.DIRECTORY:
    output = True

  for flags, regex in patterns:
    if (flags & Flags.DIRECTORY and
        not query_flags & Flags.DIRECTORY):
      continue

    if not regex.fullmatch(relpath):
      continue

    if flags & Flags.EXCLUDE:
      output = False
    else:
      most_recent_positive_match = regex
      output = True

  if output and not (query_flags & Flags.DIRECTORY):
    logger.debug("%s matched by %s", relpath, most_recent_positive_match)

  return output


def gentree(sourcedir, patterns):
  exclude = re.compile("|".join("({})".format(pat) for pat in [
      # Git directories
      r"\.git",
      r"\.gitu",
      # build artifacts
      r"\.build",
      # IDE poop
      r"\.vscode",
      # python poop
      r"__pycache__",
      # autosaves
      r".*~$",
      r"\.ipynb_checkpoints",
      # bazel poop
      "bazel-.*",
      "WORKSPACE",
      # python packaging
      r".*\.egg-info",
      "build/bdist.*",
      "build/lib",
      "dist",
      # node poop
      "node_modules",
  ]))

  for dirpath, dirnames, filenames in os.walk(sourcedir):
    relpath_parent = os.path.relpath(dirpath, sourcedir)
    if relpath_parent == ".":
      relpath_parent = ""

    filtered_dirs = []
    for dirname in sorted(dirnames):
      if exclude.fullmatch(dirname):
        continue
      reldir = os.path.join(relpath_parent, dirname)
      if not should_include(patterns, reldir, Flags.DIRECTORY):
        continue
      filtered_dirs.append(dirname)
    dirnames[:] = filtered_dirs

    for filename in sorted(filenames):
      if exclude.fullmatch(filename):
        continue
      relpath = os.path.join(relpath_parent, filename)
      if not should_include(patterns, relpath):
        continue
      yield relpath


def get_patterns_from(globspath):
  """
  Read a file containing globbing patterns and return the patterns as a list
  of strings
  """
  patterns = []
  with io.open(globspath, "r", encoding="utf-8") as infile:
    for line in infile:
      line = line.strip()
      if not line:
        continue
      if line.startswith("#"):
        continue
      patterns.append(line)
  return patterns


def translate_patterns(lines):
  """
  Translate a list of (possibly negated) globbing patterns into a list of
  regular expressions
  """
  patterns = []
  for line in lines:
    flags = 0
    if line.startswith("!"):
      line = line.lstrip("!")
      flags |= Flags.EXCLUDE
    if line.endswith("/"):
      line = line.rstrip("/")
      flags |= Flags.DIRECTORY

    # Convert shell glob to regular expression. See
    # http://man7.org/linux/man-pages/man7/glob.7.html

    # "**" matches any number of characters
    filtered_parts = []
    for part in re.split(r"((?:\*\*)|/)", line):
      if part == "**":
        filtered_parts.append(".*")
      elif part == "/":
        filtered_parts.append("/")
      else:
        # If a right square bracket is the first character in a character
        # class it is literal (e.g. part of the character class)
        part = re.sub(r"(\[!)\]", r"\1\]", part)

        # [!...] is a negated character class
        part = part.replace("[!", "[^")

        # dots are literal
        part = part.replace(".", r"\.")

        # '*' matches zero or more of anything but path separator
        part = part.replace("*", "[^/]*")

        # '?' matches zero or one of anything but path separator
        part = part.replace("?", "[^/]?")

        filtered_parts.append(part)

    translated = "".join(filtered_parts)
    patterns.append((flags, re.compile(translated)))
  return patterns


def check_manifest(treegen, sourcedir, outpath):
  latest_mtime = 0
  tmppath = outpath + ".tmp"
  with io.open(tmppath, "w", encoding="utf-8") as outfile:
    for relpath in treegen:
      outfile.write(relpath)
      outfile.write("\n")
      fullpath = os.path.join(sourcedir, relpath)
      latest_mtime = max(latest_mtime, os.path.getmtime(fullpath))

  if subprocess.call(["cmp", "--silent", outpath, tmppath]) != 0:
    print("tarball manifest is out of date")
    os.rename(tmppath, outpath)
    return

  if os.path.getmtime(outpath) < latest_mtime:
    print("tarball content is out of date")
    os.rename(tmppath, outpath)
    return


def translate_changelog(src, dst, distro):
  parent = os.path.dirname(dst)
  if not os.path.exists(parent):
    os.makedirs(parent)

  pattern = r"(\S+) \(([^\)]+)\) ubuntu; (urgency=\S+)"
  replacement = r"\1 (\2~{0}) {0}; \3".format(distro)
  entry_regex = re.compile(pattern)
  with open(src, "r") as infile:
    with open(dst, "w") as outfile:
      for line in infile:
        outfile.write(entry_regex.sub(replacement, line))


def setup_argparse(parser):
  """
  Setup argument parser
  """
  parser.add_argument(
      "--log-level", default="info",
      choices=["debug", "info", "warning", "error"])
  parser.add_argument("-o", "--outpath", help="path to output file")
  subparsers = parser.add_subparsers(dest="command")
  subparser = subparsers.add_parser(
      "parse-changelog",
      help=("Parse a debian changelog to get the package name and version"
            " numbers. Output a cmake snippet that can be include()ed"))
  subparser.add_argument("logpath")
  subparser.add_argument("prefix")

  subparser = subparsers.add_parser(
      "check-manifest",
      help=("Scan the source tree for the list of files that go into the source"
            " package. If the file set has changed, or if any of the files have"
            " changed, then touch the manifest file"))
  mutex = subparser.add_mutually_exclusive_group()
  mutex.add_argument(
      "--patterns-from", action="append", default=[],
      help="File containing a  list of patterns. Can be specified multiple"
           " times")
  mutex.add_argument(
      "--patterns", nargs="*", default=[],
      help="Patterns on the command line")
  subparser.add_argument("sourcedir")

  subparser = subparsers.add_parser(
      "translate-changelog",
      help=("Copy debian changelog and replace the most recent header line "
            "with the desired distribution"))
  subparser.add_argument("--src", help="source file")
  subparser.add_argument("--tgt", help="target/destination path")
  subparser.add_argument("--distro", help="distribution")


PARSE_CHANGELOG_FORMAT = """
set({0}_package {1})
set({0}_version {2})
set({0}_debversion {3})
"""


def parse_changelog_cmdfn(args):
  info = parse_changelog(args.logpath)
  with io.open(args.outpath, "w", encoding="utf-8") as outfile:
    outfile.write(
        PARSE_CHANGELOG_FORMAT.format(args.prefix, *info[:3]))
  return 0


def check_manifest_cmdfn(args):
  raw_patterns = list(args.patterns)
  for filepath in args.patterns_from:
    lines = get_patterns_from(filepath)
    raw_patterns.extend(lines)

  patterns = translate_patterns(raw_patterns)
  for _, pattern in patterns:
    logger.debug("%s", pattern)
  treegen = gentree(args.sourcedir, patterns)
  check_manifest(treegen, args.sourcedir, args.outpath)
  return 0


def main():
  logging.basicConfig(level=logging.INFO)
  parser = argparse.ArgumentParser(description=__doc__)
  setup_argparse(parser)
  args = parser.parse_args()
  logging.getLogger().setLevel(getattr(logging, args.log_level.upper()))

  if args.command == "parse-changelog":
    return parse_changelog_cmdfn(args)
  elif args.command == "check-manifest":
    return check_manifest_cmdfn(args)
  elif args.command == "translate-changelog":
    translate_changelog(args.src, args.tgt, args.distro)
  else:
    logging.error("Unknown command: {}".format(args.command))
    return 1

  return 0


if __name__ == "__main__":
  logging.basicConfig(level=logging.INFO, format="%(levelname)s %(message)s")
  sys.exit(main())


#
#
#
#
#
#
#
#
#
#
#
#
#
#


def get_progress_bar(fraction, numchars=30):
  """
  Return a high resolution unicode progress bar
  """
  blocks = ["", "▏", "▎", "▍", "▌", "▋", "▊", "▉", "█"]
  length_in_chars = fraction * numchars
  if length_in_chars > numchars:
    length_in_chars = numchars
  n_full = int(length_in_chars)

  i_partial = int(len(blocks) * (length_in_chars - n_full))
  partial = blocks[i_partial]
  n_empty = max(numchars - n_full - len(partial), 0)
  return ("█" * n_full) + partial + (" " * n_empty)


def get_human_readable_size(size_in_bytes):
  """
  Convert a number of bytes into a human readable string.
  """
  if size_in_bytes == 0:
    return '{:6.2f}{}'.format(0, " B")

  exponent = int(math.log(size_in_bytes, 1024))
  unit = [' B', 'KB', 'MB', 'GB', 'PB', 'EB'][exponent]
  size_in_units = float(size_in_bytes) / (1024 ** exponent)
  return '{:6.2f}{}'.format(size_in_units, unit)


def download_file(srcurl, outpath):
  """
  Download <package>-<version>.tar.gz if it doesn't already exist
  """
  tmppath = outpath + ".tmp"
  if os.path.exists(tmppath):
    os.remove(tmppath)

  response = requests.head(srcurl)
  totalsize = int(response.headers.get("content-length", 2320000))
  recvsize = 0

  request = requests.get(srcurl, stream=True)
  last_print = 0

  outname = os.path.basename(outpath)
  with open(tmppath, "wb") as outfile:
    for chunk in request.iter_content(chunk_size=4096):
      outfile.write(chunk)
      recvsize += len(chunk)

      if time.time() - last_print > 0.1:
        last_print = time.time()
        percent = 100.0 * recvsize / totalsize
        message = ("Downloading {}: {}/{} [{}] {:6.2f}%"
                   .format(outname,
                           get_human_readable_size(recvsize),
                           get_human_readable_size(totalsize),
                           get_progress_bar(percent / 100.0), percent))
        sys.stdout.write(message)
        sys.stdout.flush()
        sys.stdout.write("\r")
  message = ("Downloading {}: {}/{} [{}] {:6.2f}%"
             .format(outname,
                     get_human_readable_size(totalsize),
                     get_human_readable_size(totalsize),
                     get_progress_bar(1.0), 100.0))
  sys.stdout.write(message)
  sys.stdout.write("\n")
  sys.stdout.flush()
  os.rename(tmppath, outpath)
  return outpath


def get_base_tgz(distro, arch):
  return "/var/cache/pbuilder/{}-{}-base.tgz".format(distro, arch)


def prep_pbuilder(distro, arch, basetgz):
  """
  Create base rootfs tarfiles for the specified distro/arch.
  """
  subprocess.check_call(
      ["sudo", "pbuilder", "--create", "--distribution", distro,
       "--architecture", arch, "--basetgz", basetgz])


def exec_pbuilder(dscpath, distro, arch, basetgz, outdir):
  """
  Execute pbuilder to get the binary archives
  """
  subprocess.check_call(
      ["sudo", "env", 'DEB_BUILD_OPTIONS="parallel=8"',
       "pbuilder", "--build", "--distribution", distro,
       "--architecture", arch, "--basetgz", basetgz,
       "--buildresult", outdir, dscpath])


def build_arch(distro, arch, args, dscpath, package_name, upstream_version,
               local_version):
  workdir = os.path.join(args.out, distro)

  # Create the base tarball if needed
  basetgz = get_base_tgz(distro, arch)
  outname = os.path.basename(basetgz)
  if os.path.exists(basetgz):
    logger.info("%s up to date", outname)
  else:
    logger.info("Making %s", outname)
    prep_pbuilder(distro, arch, basetgz)

  outname = ("{}_{}-{}_{}.deb"
             .format(package_name, upstream_version, local_version, arch))

  binout = os.path.join(workdir, 'binary')
  if not os.path.exists(binout):
    os.makedirs(binout)

  outpath = os.path.join(binout, outname)
  needs_build = False
  if os.path.exists(outpath):
    if os.stat(outpath).st_mtime < os.stat(dscpath).st_mtime:
      needs_build = True
      logger.info("%s is out of date", outname)
    else:
      logger.info("%s is up to date", outname)
  else:
    needs_build = True
    logger.info("Need to create %s", outname)

  if needs_build:
    exec_pbuilder(dscpath, distro, arch, basetgz, binout)

  logger.info("Writing out repository index")
  with open(os.path.join(binout, "Packages.gz"), "wb") as outfile:
    scanpkg = subprocess.Popen(
        ["dpkg-scanpackages", "."], cwd=binout, stdout=subprocess.PIPE)
    gzip = subprocess.Popen(
        ["gzip", "-9c"], cwd=binout, stdout=outfile, stdin=scanpkg.stdout)
    scanpkg.stdout.close()
  scanpkg.wait()
  gzip.wait()


def build_for(distro, args, tarball):
  logger.info("Building for %s", distro)
  workdir = os.path.join(args.out, distro)
  if not os.path.exists(workdir):
    os.makedirs(workdir)

  # Extract the tarball
  srcdir = os.path.join(workdir, "i3-{}".format(UPSTREAM_VERSION))
  if os.path.exists(srcdir):
    logger.info("Already extracted tarball")
  else:
    logger.info("Extracting tarball")
    subprocess.check_call(["tar", "xf", tarball], cwd=workdir)

  # Create symlink to source
  tarname = os.path.basename(tarball)
  tarlink = os.path.join(workdir, tarname)
  if not os.path.lexists(tarlink):
    os.symlink(tarball, tarlink)

  # Synchronize patches
  logger.info("Syncing debian patches")
  subprocess.check_call([
      'rsync', "-a",
      os.path.join(args.src, "debian/"),
      os.path.join(srcdir, "debian/")
  ])

  # Translate the changelog into ubuntu suite names
  src_changelog = os.path.join(args.src, "debian/changelog")
  srcdir = os.path.join(workdir, "i3-{}".format(UPSTREAM_VERSION))
  dst_changelog = os.path.join(srcdir, "debian/changelog")
  translate_changelog(src_changelog, dst_changelog, distro)

  # Create the .dsc file
  (package_name, upstream_version, local_version,
      distribution) = parse_changelog(dst_changelog)
  if distro != distribution:
    raise RuntimeError("{} != {}".format(distro, distribution))

  outname = ("{}_{}-{}.dsc"
             .format(package_name, upstream_version, local_version))
  dscpath = os.path.join(workdir, outname)
  if os.path.exists(dscpath):
    logger.info("%s already built", outname)
  else:
    logger.info("Creating %s", outname)
    subprocess.check_call(['debuild', "-S", "-sa", "-pgpg2", "-k6A8A4FAF"],
                          cwd=srcdir)

  if args.skip_build:
    return
  for arch in args.arch:
    build_arch(distro, arch, args, dscpath, package_name, upstream_version,
               local_version)


def download_if_needed(src_url, out):
   # Download the tarball if needed
  if src_url:
    filename = src_url.rsplit("/", 1)[1]
    basename = filename[:-len(".tar.gz")]
    tarball = os.path.join(out, basename + ".orig.tar.gz")
    if not os.path.exists(out):
      os.makedirs(out)
    if os.path.exists(tarball):
      logger.info("Already downloaded tarball")
    else:
      download_file(src_url, tarball)



"""
Generate the rtd-requirements.txt file given a release tag and version number
"""

import argparse
import io
import os
import sys


def main():
  parser = argparse.ArgumentParser(description=__doc__)
  parser.add_argument("--tag", required=True)
  parser.add_argument("--version", required=True)
  parser.add_argument("-o", "--outfile-path", default="-")
  parser.add_argument("template")

  args = parser.parse_args()
  if args.tag == "from-travis":
    args.tag = os.environ.get("TRAVIS_TAG", "pseudo-staging")

  if args.outfile_path == "-":
    args.outfile_path = os.dup(sys.stdout.fileno())

  outfile = io.open(args.outfile_path, "w", encoding="utf-8")
  with io.open(args.template, "r", encoding="utf-8") as infile:
    template = infile.read()
  content = template.format(_tag=args.tag, _version=args.version)
  outfile.write(content)


if __name__ == "__main__":
  main()



from __future__ import unicode_literals
import collections
import logging

from cmakelang.lint import lintdb

logger = logging.getLogger(__name__)


class LintRecord(object):
  """Records an instance of lint at a particular location
  """

  def __init__(self, spec, location, msg):
    self.spec = spec
    self.location = location
    self.msg = msg

  def __repr__(self):
    if self.location is None:
      return " [{:s}] {:s}".format(self.spec.idstr, self.msg)

    if isinstance(self.location, tuple):
      return "{:s}: [{:s}] {:s}".format(
          ",".join("{:02d}".format(val) for val in self.location[:2]),
          self.spec.idstr, self.msg)

    raise ValueError(
        "Unexpected type {} for location".format(type(self.location)))


SuppressionEvent = collections.namedtuple(
    "SuppresionEvent", ["lineno", "mode", "suppressions"])


class FileContext(object):
  def __init__(self, global_ctx, infile_path):
    self.global_ctx = global_ctx
    self.infile_path = infile_path
    self.config = None
    self._lint = []

    # Suppressions active at the current depth
    self._suppressions = set()
    self._supressed_count = collections.defaultdict(int)

    # Map of line number to list of active suppressions starting at this line
    # and ending at the line number of the next entry in the list
    self._suppression_events = []

  def is_idstr(self, idstr):
    return idstr in self.global_ctx.lintdb

  def suppress(self, lineno, idlist):
    """
    Given a list of lint ids, enable a suppression for each one which is not
    already supressed. Return the list of new suppressions
    """
    new_suppressions = []
    for idstr in idlist:
      if idstr in self._suppressions:
        continue
      self._suppressions.add(idstr)
      new_suppressions.append(idstr)

    self._suppression_events.append(
        SuppressionEvent(lineno, "add", list(new_suppressions)))
    return new_suppressions

  def unsuppress(self, lineno, idlist):
    for idstr in idlist:
      if idstr not in self._suppressions:
        logger.warning(
            "Unsupressing %s which is not currently surpressed", idstr)
      self._suppressions.discard(idstr)
    self._suppression_events.append(
        SuppressionEvent(lineno, "remove", list(idlist)))

  def record_lint(self, idstr, *args, **kwargs):
    if idstr in self.config.lint.disabled_codes:
      self._supressed_count[idstr] += 1
      return

    if idstr in self._suppressions:
      self._supressed_count[idstr] += 1
      return

    spec = self.global_ctx.lintdb[idstr]
    location = kwargs.pop("location", ())
    msg = spec.msgfmt.format(*args, **kwargs)
    record = LintRecord(spec, location, msg)
    self._lint.append(record)

  def get_lint(self):
    """Return lint records in sorted order"""
    records = [
        record for _, _, _, record in sorted(
            (record.location, record.spec.idstr, idx, record)
            for idx, record in enumerate(self._lint))]

    # Remove any lint records that were suppressed at the line number where
    # they were recorded
    out = []
    events = list(self._suppression_events)
    active_suppressions = set()

    for record in records:
      if record.location:
        while events and record.location[0] >= events[0].lineno:
          event = events.pop(0)
          if event.mode == "add":
            for idstr in event.suppressions:
              active_suppressions.add(idstr)
          elif event.mode == "remove":
            for idstr in event.suppressions:
              active_suppressions.discard(idstr)
          else:
            raise ValueError("Illegal suppression event {}".format(event.mode))
      if record.spec.idstr in active_suppressions:
        continue
      out.append(record)

    return out

  def writeout(self, outfile):
    for record in self.get_lint():
      outfile.write("{:s}:{}\n".format(self.infile_path, record))

  def has_lint(self):
    return bool(self.get_lint())


class GlobalContext(object):
  category_names = {
      "C": "Convention",
      "E": "Error",
      "R": "Refactor",
      "W": "Warning",
  }

  def __init__(self, outfile):
    self.outfile = outfile
    self.lintdb = lintdb.get_database()
    self.file_ctxs = {}

  def get_file_ctx(self, infile_path, config):
    if infile_path not in self.file_ctxs:
      self.file_ctxs[infile_path] = FileContext(self, infile_path)
    ctx = self.file_ctxs[infile_path]
    ctx.config = config
    return ctx

  def get_category_counts(self):
    lint_counts = {}
    for _, file_ctx in sorted(self.file_ctxs.items()):
      for record in file_ctx.get_lint():
        category_char = record.spec.idstr[0]
        if category_char not in lint_counts:
          lint_counts[category_char] = 0
        lint_counts[category_char] += 1
    return lint_counts

  def write_summary(self, outfile):
    outfile.write("Summary\n=======\n")
    outfile.write("files scanned: {:d}\n".format(len(self.file_ctxs)))
    outfile.write("found lint:\n")

    lint_counts = self.get_category_counts()
    fieldwidth = max(len(name) for _, name in self.category_names.items())
    fmtstr = "  {:>" + str(fieldwidth) + "s}: {:d}\n"
    for (category_char, count) in sorted(lint_counts.items()):
      category_name = self.category_names[category_char]
      outfile.write(fmtstr.format(category_name, count))
    outfile.write("\n")



# pylint: disable=too-many-statements
import logging

from cmakelang.lex import TokenType
from cmakelang.parse.additional_nodes import (
    FlagGroupNode, ShellCommandNode, TupleParser)
from cmakelang.parse.argument_nodes import (
    ArgGroupNode,
    KeywordGroupNode,
    PositionalGroupNode,
    PositionalParser,
    StandardArgTree)
from cmakelang.parse.common import KwargBreaker, NodeType, TreeNode
from cmakelang.parse.util import (
    WHITESPACE_TOKENS,
    get_first_semantic_token,
    get_normalized_kwarg,
    should_break)

logger = logging.getLogger(__name__)


def parse_add_custom_command_events(ctx, tokens, breakstack):
  """
  ::
    add_custom_command(TARGET <target>
                    PRE_BUILD | PRE_LINK | POST_BUILD
                    COMMAND command1 [ARGS] [args1...]
                    [COMMAND command2 [ARGS] [args2...] ...]
                    [BYPRODUCTS [files...]]
                    [WORKING_DIRECTORY dir]
                    [COMMENT comment]
                    [VERBATIM] [USES_TERMINAL]
                    [COMMAND_EXPAND_LISTS])
  :see: https://cmake.org/cmake/help/latest/command/add_custom_command.html
  """
  subtree = StandardArgTree.parse(
      ctx, tokens,
      npargs='*',
      kwargs={
          "BYPRODUCTS": PositionalParser('*'),
          "COMMAND": ShellCommandNode.parse,
          "COMMENT": PositionalParser('*'),
          "TARGET": PositionalParser(1),
          "WORKING_DIRECTORY": PositionalParser(1)
      },
      flags=[
          "APPEND", "VERBATIM", "USES_TERMINAL", "COMMAND_EXPAND_LISTS",
          "PRE_BUILD", "PRE_LINK", "POST_BUILD"],
      breakstack=breakstack)

  subtree.check_required_kwargs(ctx.lint_ctx, {
      # Truly required keyword arguments
      "COMMAND": "E1125",
      # Required by convention
      "COMMENT": "C0113"
  })
  return subtree


def parse_add_custom_command_standard(ctx, tokens, breakstack):
  """
  ::

      add_custom_command(OUTPUT output1 [output2 ...]
                         COMMAND command1 [ARGS] [args1...]
                         [COMMAND command2 [ARGS] [args2...] ...]
                         [MAIN_DEPENDENCY depend]
                         [DEPENDS [depends...]]
                         [BYPRODUCTS [files...]]
                         [IMPLICIT_DEPENDS <lang1> depend1
                                          [<lang2> depend2] ...]
                         [WORKING_DIRECTORY dir]
                         [COMMENT comment]
                         [DEPFILE depfile]
                         [JOB_POOL job_pool]
                         [VERBATIM] [APPEND] [USES_TERMINAL]
                         [COMMAND_EXPAND_LISTS])

  :see: https://cmake.org/cmake/help/latest/command/add_custom_command.html
  """
  subtree = StandardArgTree.parse(
      ctx, tokens,
      npargs='*',
      kwargs={
          "BYPRODUCTS": PositionalParser('*'),
          "COMMAND": ShellCommandNode.parse,
          "COMMENT": PositionalParser('*'),
          "DEPENDS": PositionalParser('*'),
          "DEPFILE": PositionalParser(1),
          "JOB_POOL": PositionalParser(1),
          "IMPLICIT_DEPENDS": TupleParser(2, '+'),
          "MAIN_DEPENDENCY": PositionalParser(1),
          "OUTPUT": PositionalParser('+'),
          "WORKING_DIRECTORY": PositionalParser(1)
      },
      flags=["APPEND", "VERBATIM", "USES_TERMINAL", "COMMAND_EXPAND_LISTS"],
      breakstack=breakstack)

  subtree.check_required_kwargs(ctx.lint_ctx, {
      # Truly required keyword arguments
      "COMMAND": "E1125",
      # Required by convention
      "COMMENT": "C0113"
  })

  return subtree


def parse_add_custom_command(ctx, tokens, breakstack):
  """
  There are two forms of `add_custom_command`. This is the dispatcher between
  the two forms.
  """
  descriminator_token = get_first_semantic_token(tokens)
  if (descriminator_token is None or
      descriminator_token.type is TokenType.RIGHT_PAREN):
    location = ()
    if tokens:
      location = tokens[0].get_location()
    # logger.warning("Invalid empty file() command at %s", location)
    ctx.lint_ctx.record_lint("E1120", location=location)
    return StandardArgTree.parse(ctx, tokens, npargs='*', kwargs={}, flags=[],
                                 breakstack=breakstack)

  if descriminator_token.type is TokenType.DEREF:
    ctx.lint_ctx.record_lint(
        "C0114", location=descriminator_token.get_location())
    return parse_add_custom_command_standard(ctx, tokens, breakstack)

  descriminator_word = get_normalized_kwarg(descriminator_token)
  if descriminator_word == "TARGET":
    return parse_add_custom_command_events(ctx, tokens, breakstack)
  if descriminator_word == "OUTPUT":
    return parse_add_custom_command_standard(ctx, tokens, breakstack)

  # logger.warning(
  #     "Indeterminate form of add_custom_command \"%s\" at %s",
  #     descriminator_word, descriminator_token.location())
  ctx.lint_ctx.record_lint("E1126", location=descriminator_token.get_location())
  return parse_add_custom_command_standard(ctx, tokens, breakstack)


def parse_add_custom_target(ctx, tokens, breakstack):
  """
  ::
    add_custom_target(Name [ALL] [command1 [args1...]]
                      [COMMAND command2 [args2...] ...]
                      [DEPENDS depend depend depend ... ]
                      [BYPRODUCTS [files...]]
                      [WORKING_DIRECTORY dir]
                      [COMMENT comment]
                      [JOB_POOL job_pool]
                      [VERBATIM] [USES_TERMINAL]
                      [COMMAND_EXPAND_LISTS]
                      [SOURCES src1 [src2...]])

  :see: https://cmake.org/cmake/help/latest/command/add_custom_target.html
  """
  kwargs = {
      "BYPRODUCTS": PositionalParser("+"),
      "COMMAND": ShellCommandNode.parse,
      "COMMENT": PositionalParser(1),
      "DEPENDS": PositionalParser("+"),
      "JOB_POOL": PositionalParser(1),
      "SOURCES": PositionalParser("+"),
      "WORKING_DIRECTORY": PositionalParser(1),
  }

  required_kwargs = {
      # Required by convention
      "COMMENT": "C0113"
  }

  flags = ("VERBATIM", "USES_TERMINAL", "COMMAND_EXPAND_LISTS")
  tree = ArgGroupNode()

  # If it is a whitespace token then put it directly in the parse tree at
  # the current depth
  while tokens and tokens[0].type in WHITESPACE_TOKENS:
    tree.children.append(tokens.pop(0))
    continue

  breaker = KwargBreaker(list(kwargs.keys()) + list(flags))
  child_breakstack = breakstack + [breaker]

  nametree = None
  state = "name"

  while tokens:
    # Break if the next token belongs to a parent parser, i.e. if it
    # matches a keyword argument of something higher in the stack, or if
    # it closes a parent group.
    if should_break(tokens[0], breakstack):
      break

    # If it is a whitespace token then put it directly in the parse tree at
    # the current depth
    if tokens[0].type in WHITESPACE_TOKENS:
      tree.children.append(tokens.pop(0))
      continue

    # If it's a comment, then add it at the current depth
    if tokens[0].type in (TokenType.COMMENT,
                          TokenType.BRACKET_COMMENT):
      child = TreeNode(NodeType.COMMENT)
      tree.children.append(child)
      child.children.append(tokens.pop(0))
      continue

    ntokens = len(tokens)
    if state == "name":
      next_semantic = get_first_semantic_token(tokens[1:])
      if (next_semantic is not None and
          get_normalized_kwarg(next_semantic) == "ALL"):
        npargs = 2
      else:
        npargs = 1

      nametree = PositionalGroupNode.parse(
          ctx, tokens, npargs, ["ALL"], child_breakstack)
      assert len(tokens) < ntokens
      tree.children.append(nametree)
      state = "first-command"
      continue

    word = get_normalized_kwarg(tokens[0])
    if state == "first-command":
      if not(word in kwargs or word in flags):
        subtree = PositionalGroupNode.parse(
            ctx, tokens, '+', [], child_breakstack)
        tree.children.append(subtree)
        assert len(tokens) < ntokens
      state = "kwargs"
      continue

    if word in flags:
      subtree = FlagGroupNode.parse(
          ctx, tokens, flags, breakstack + [KwargBreaker(list(kwargs.keys()))])
      assert len(tokens) < ntokens
      tree.children.append(subtree)
      continue

    if word in kwargs:
      required_kwargs.pop(word, None)
      subtree = KeywordGroupNode.parse(
          ctx, tokens, word, kwargs[word], child_breakstack)
      assert len(tokens) < ntokens
      tree.children.append(subtree)
      continue

    ctx.lint_ctx.record_lint("E1122", location=tokens[0].get_location())
    # logger.warning(
    #     "Unexpected positional argument %s at %s",
    #     tokens[0].spelling, tokens[0].location())
    subtree = PositionalGroupNode.parse(ctx, tokens, '+', [], child_breakstack)
    assert len(tokens) < ntokens
    tree.children.append(subtree)
    continue

  if required_kwargs:
    location = ()
    for token in tree.get_semantic_tokens():
      location = token.get_location()
      break

    missing_kwargs = sorted(
        (lintid, word) for word, lintid in sorted(required_kwargs.items()))
    for lintid, word in missing_kwargs:
      ctx.lint_ctx.record_lint(lintid, word, location=location)
  return tree


def parse_add_compile_options(ctx, tokens, breakstack):
  """
  ::

    add_compile_options(<option> ...)

  :see: https://cmake.org/cmake/help/latest/command/add_compile_options.html
  """
  return StandardArgTree.parse(ctx, tokens, "+", {}, [], breakstack)


def parse_add_definitions(ctx, tokens, breakstack):
  """
  ::

    add_definitions(-DFOO -DBAR ...)

  :see: https://cmake.org/cmake/help/latest/command/add_definitions.html
  """
  return StandardArgTree.parse(ctx, tokens, "+", {}, [], breakstack)


def parse_add_dependencies(ctx, tokens, breakstack):
  """
  ::

    add_dependencies(<target> [<target-dependency>]...)

  :see: https://cmake.org/cmake/help/latest/command/add_dependencies.html
  """

  return StandardArgTree.parse(ctx, tokens, "2+", {}, [], breakstack)


def parse_add_subdirectory(ctx, tokens, breakstack):
  """
  ::

    add_subdirectory(source_dir [binary_dir] [EXCLUDE_FROM_ALL])

  :see: https://cmake.org/cmake/help/latest/command/add_subdirectory.html
  """

  return StandardArgTree.parse(
      ctx, tokens, "+", {}, ["EXCLUDE_FROM_ALL"], breakstack)


def parse_aux_source_directory(ctx, tokens, breakstack):
  """
  ::

    aux_source_directory(<dir> <variable>)

  :see: https://cmake.org/cmake/help/latest/command/aux_source_directory.html
  """

  return StandardArgTree.parse(
      ctx, tokens, "2", {}, [], breakstack)


def populate_db(parse_db):
  parse_db["add_custom_command"] = parse_add_custom_command
  parse_db["add_custom_target"] = parse_add_custom_target
  parse_db["add_compile_options"] = parse_add_compile_options
  parse_db["add_definitions"] = parse_add_definitions
  parse_db["add_dependencies"] = parse_add_dependencies
  parse_db["add_subdirectory"] = parse_add_subdirectory
  parse_db["aux_source_directory"] = parse_aux_source_directory



from cmakelang import lex
from cmakelang.parse.argument_nodes import (
    PositionalParser,
    StandardArgTree,
)
from cmakelang.parse.additional_nodes import (
    TupleParser
)
from cmakelang.parse.util import (
    get_nth_semantic_token
)


def parse_build_command(ctx, tokens, breakstack):
  """
  ::

    build_command(<variable>
                  [CONFIGURATION <config>]
                  [TARGET <target>]
                  [PROJECT_NAME <projname>] # legacy, causes warning
                 )

  :see: https://cmake.org/cmake/help/latest/command/build_command.html
  """
  kwargs = {
      "CONFIGURATION": PositionalParser(1),
      "TARGET": PositionalParser(1),
      "PROJECT_NAME": PositionalParser(1)
  }

  second_token = get_nth_semantic_token(tokens, 1)
  if second_token is not None and second_token.spelling.upper not in kwargs:
    ctx.lint_ctx.record_lint(
        "W0106", "build_command(<cachevariable> <makecommand>)",
        location=tokens[0].get_location())
    return StandardArgTree.parse(ctx, tokens, "2", {}, [], breakstack)

  return StandardArgTree.parse(ctx, tokens, "1", kwargs, [], breakstack)


def parse_cmake_parse_arguments(ctx, tokens, breakstack):
  """
  ::

  cmake_parse_arguments(<prefix> <options> <one_value_keywords>
                      <multi_value_keywords> <args>...)

  cmake_parse_arguments(PARSE_ARGV <N> <prefix> <options>
                        <one_value_keywords> <multi_value_keywords>)

  :see: https://cmake.org/cmake/help/latest/command/cmake_parse_arguments.html
  """

  first_token = get_nth_semantic_token(tokens, 0)
  if first_token is not None and first_token.spelling.upper == "PARSE_ARGV":
    return StandardArgTree.parse(
        ctx, tokens, "6", {}, ["PARSE_ARGV"], breakstack)
  return StandardArgTree.parse(
      ctx, tokens, "5+", {}, [], breakstack)


def parse_cmake_policy(ctx, tokens, breakstack):
  """
  ::

    cmake_policy(VERSION <min>[...<max>])
    cmake_policy(SET CMP<NNNN> NEW)
    cmake_policy(SET CMP<NNNN> OLD)
    cmake_policy(GET CMP<NNNN> <variable>)
    cmake_policy(PUSH)
    cmake_policy(POP)

  :see: https://cmake.org/cmake/help/latest/command/cmake_policy.html
  """
  first_token = get_nth_semantic_token(tokens, 0)
  if first_token is None:
    return StandardArgTree.parse(ctx, tokens, "+", {}, [], breakstack)

  if first_token.type is lex.TokenType.DEREF:
    ctx.lint_ctx.record_lint("C0114", location=first_token.get_location())
    return StandardArgTree.parse(ctx, tokens, "+", {}, [], breakstack)

  descriminator = first_token.spelling.upper()
  if descriminator == "VERSION":
    return StandardArgTree.parse(ctx, tokens, "2+", {}, ["VERSION"], breakstack)

  if descriminator == "SET":
    return StandardArgTree.parse(
        ctx, tokens, 3, {}, ["SET", "OLD", "NEW"], breakstack)

  if descriminator == "GET":
    return StandardArgTree.parse(ctx, tokens, 3, {}, ["GET"], breakstack)

  if descriminator in ("PUSH", "POP"):
    return StandardArgTree.parse(
        ctx, tokens, 1, {}, ["PUSH", "POP"], breakstack)

  ctx.lint_ctx.record_lint("E1126", location=first_token.get_location())
  return StandardArgTree.parse(
      ctx, tokens, 1, {}, ["PUSH", "POP"], breakstack)


def parse_create_test_sourcelist(ctx, tokens, breakstack):
  """
  ::

    create_test_sourcelist(sourceListName driverName
                           test1 test2 test3
                           EXTRA_INCLUDE include.h
                           FUNCTION function)

  :see: https://cmake.org/cmake/help/latest/command/create_test_sourcelist.html
  """

  kwargs = {
      "EXTRA_INCLUDE": PositionalParser(1),
      "FUNCTION": PositionalParser(1)
  }
  return StandardArgTree.parse(
      ctx, tokens, "3+", kwargs, [], breakstack)


def parse_export_library_dependencies(ctx, tokens, breakstack):
  """
  ::

    export_library_dependencies(<file> [APPEND])

  :see: https://cmake.org/cmake/help/latest/command/export_library_dependencies.html
  """
  return StandardArgTree.parse(
      ctx, tokens, "1+", {}, ["APPEND"], breakstack)


def parse_fltk_wrap_ui(ctx, tokens, breakstack):
  """
  ::

      fltk_wrap_ui(resultingLibraryName source1
                   source2 ... sourceN )

  :see: https://cmake.org/cmake/help/latest/command/fltk_wrap_ui.html
  """
  return StandardArgTree.parse(
      ctx, tokens, "2+", {}, [], breakstack)


def parse_get_cmake_property(ctx, tokens, breakstack):
  """
  ::

      get_cmake_property(VAR property)

  :see: https://cmake.org/cmake/help/latest/command/get_cmake_property.html
  """
  return StandardArgTree.parse(ctx, tokens, 2, {}, [], breakstack)


def parse_get_filename_component(ctx, tokens, breakstack):
  """
  ::

      get_filename_component(<VAR> <FileName> <COMP> [CACHE])

      get_filename_component(<VAR> FileName
                             PROGRAM [PROGRAM_ARGS <ARG_VAR>]
                             [CACHE])

  :see: https://cmake.org/cmake/help/latest/command/get_filename_component.html
  """
  descriminator = get_nth_semantic_token(tokens, 2)
  if descriminator is not None and descriminator.spelling.upper() == "PROGRAM":
    flags = ["PROGRAM", "PROGRAM_ARGS", "CACHE"]
    return StandardArgTree.parse(ctx, tokens, "3+", {}, flags, breakstack)

  flags = [
      "DIRECTORY",
      "NAME",
      "EXT",
      "NAME_WE",
      "ABSOLUTE",
      "REALPATH",
      "PATH",
  ]
  return StandardArgTree.parse(ctx, tokens, "3+", {}, flags, breakstack)


def parse_get_source_file_property(ctx, tokens, breakstack):
  """
  ::

    get_source_file_property(VAR file property)

  :see: https://cmake.org/cmake/help/latest/command/get_source_file_property.html
  """
  return StandardArgTree.parse(ctx, tokens, 3, {}, [], breakstack)


def parse_get_target_property(ctx, tokens, breakstack):
  """
  ::

    get_target_property(VAR target property)

  :see: https://cmake.org/cmake/help/latest/command/get_target_property.html
  """
  return StandardArgTree.parse(ctx, tokens, 3, {}, [], breakstack)


def parse_get_test_property(ctx, tokens, breakstack):
  """
  ::

    get_test_property(test property VAR)

  :see: https://cmake.org/cmake/help/latest/command/get_test_property.html
  """
  return StandardArgTree.parse(ctx, tokens, 3, {}, [], breakstack)


def parse_include_external_msproject(ctx, tokens, breakstack):
  """
  ::

    include_external_msproject(projectname location
                           [TYPE projectTypeGUID]
                           [GUID projectGUID]
                           [PLATFORM platformName]
                           dep1 dep2 ...)

  :see: https://cmake.org/cmake/help/latest/command/include_external_msproject.html
  """
  kwargs = {
      "TYPE": PositionalParser(1),
      "GUID": PositionalParser(1),
      "PLATFORM": PositionalParser(1)
  }
  return StandardArgTree.parse(ctx, tokens, "2+", kwargs, [], breakstack)


def parse_include_guard(ctx, tokens, breakstack):
  """
  ::

    include_guard([DIRECTORY|GLOBAL])

  :see: https://cmake.org/cmake/help/latest/command/include_guard.html
  """
  return StandardArgTree.parse(ctx, tokens, "*", {}, [], breakstack)


def parse_include_regular_expression(ctx, tokens, breakstack):
  """
  ::

    include_regular_expression(regex_match [regex_complain])

  :see: https://cmake.org/cmake/help/latest/command/include_regular_expression.html
  """
  return StandardArgTree.parse(ctx, tokens, "*", {}, [], breakstack)


def parse_link_directories(ctx, tokens, breakstack):
  """
  ::

    link_directories([AFTER|BEFORE] directory1 [directory2 ...])

  :see: https://cmake.org/cmake/help/latest/command/link_directories.html
  """
  return StandardArgTree.parse(
      ctx, tokens, "+", {}, ["AFTER", "BEFORE"], breakstack)


def parse_link_libraries(ctx, tokens, breakstack):
  """
  ::

    link_libraries([item1 [item2 [...]]]
                   [[debug|optimized|general] <item>] ...)

  :see: https://cmake.org/cmake/help/latest/command/link_libraries.html
  """
  return StandardArgTree.parse(
      ctx, tokens, "+", {}, ["DEBUG", "OPTIMIZED", "GENERAL"], breakstack)


def parse_load_cache(ctx, tokens, breakstack):
  """
  ::

    load_cache(pathToBuildDirectory READ_WITH_PREFIX prefix entry1...)
    load_cache(pathToBuildDirectory [EXCLUDE entry1...]
               [INCLUDE_INTERNALS entry1...])

  :see: https://cmake.org/cmake/help/latest/command/load_cache.html
  """
  kwargs = {
      "READ_WITH_PREFIX": PositionalParser("2+"),
      "EXCLUDE": PositionalParser("+"),
      "INCLUDE_INTERNALS": PositionalParser("+")
  }
  return StandardArgTree.parse(ctx, tokens, 1, kwargs, [], breakstack)


def parse_load_command(ctx, tokens, breakstack):
  """
  ::

    load_command(COMMAND_NAME <loc1> [loc2 ...])

  :see: https://cmake.org/cmake/help/latest/command/load_command.html
  """
  return StandardArgTree.parse(
      ctx, tokens, "2+", {}, ["COMMAND_NAME"], breakstack)


def parse_math(ctx, tokens, breakstack):
  """
  ::

    math(EXPR <variable> "<expression>" [OUTPUT_FORMAT <format>])

  :see: https://cmake.org/cmake/help/latest/command/math.html
  """
  kwargs = {
      "OUTPUT_FORMAT": PositionalParser(1)
  }
  return StandardArgTree.parse(ctx, tokens, "3", kwargs, [], breakstack)


def parse_option(ctx, tokens, breakstack):
  """
  ::

    option(<variable> "<help_text>" [value])

  :see: https://cmake.org/cmake/help/latest/command/option.html
  """
  return StandardArgTree.parse(ctx, tokens, "2+", {}, [], breakstack)


def parse_output_required_files(ctx, tokens, breakstack):
  """
  ::

    output_required_files(srcfile outputfile)

  :see: https://cmake.org/cmake/help/latest/command/output_required_files.html
  """
  return StandardArgTree.parse(ctx, tokens, 2, {}, [], breakstack)


def parse_qt_wrap_cpp(ctx, tokens, breakstack):
  """
  ::

    qt_wrap_cpp(resultingLibraryName DestName SourceLists ...)

  :see: https://cmake.org/cmake/help/latest/command/qt_wrap_cpp.html
  """
  return StandardArgTree.parse(ctx, tokens, "3+", {}, [], breakstack)


def parse_qt_wrap_ui(ctx, tokens, breakstack):
  """
  ::

    qt_wrap_ui(resultingLibraryName HeadersDestName
               SourcesDestName SourceLists ...)

  :see: https://cmake.org/cmake/help/latest/command/qt_wrap_ui.html
  """
  return StandardArgTree.parse(ctx, tokens, "4+", {}, [], breakstack)


def parse_remove_definitions(ctx, tokens, breakstack):
  """
  ::

    remove_definitions(-DFOO -DBAR ...)

  :see: https://cmake.org/cmake/help/latest/command/remove_definitions.html
  """
  return StandardArgTree.parse(ctx, tokens, "+", {}, [], breakstack)


def parse_separate_arguments(ctx, tokens, breakstack):
  """
  ::

    separate_arguments(<variable> <mode> <args>)

  :see: https://cmake.org/cmake/help/latest/command/separate_arguments.html
  """
  flags = ["UNIX_COMMAND", "WINDOWS_COMMAND", "NATIVE_COMMAND"]
  return StandardArgTree.parse(ctx, tokens, 3, {}, flags, breakstack)


def parse_set_source_files_properties(ctx, tokens, breakstack):
  """
  ::

    set_source_files_properties([file1 [file2 [...]]]
                                PROPERTIES prop1 value1
                                [prop2 value2 [...]])

  :see: https://cmake.org/cmake/help/latest/command/set_source_files_properties.html
  """
  kwargs = {
      "PROPERTIES": TupleParser(2, "+")
  }
  return StandardArgTree.parse(ctx, tokens, "+", kwargs, [], breakstack)


def parse_site_name(ctx, tokens, breakstack):
  """
  ::

    site_name(variable)

  :see: https://cmake.org/cmake/help/latest/command/site_name.html
  """
  return StandardArgTree.parse(ctx, tokens, 1, {}, [], breakstack)


def parse_source_group(ctx, tokens, breakstack):
  """
  ::

    source_group(<name> [FILES <src>...] [REGULAR_EXPRESSION <regex>])
    source_group(TREE <root> [PREFIX <prefix>] [FILES <src>...])
    source_group(<name> <regex>)

  :see: https://cmake.org/cmake/help/latest/command/source_group.html
  """
  descriminator = get_nth_semantic_token(tokens, 0)
  if descriminator is not None and descriminator.spelling.upper() == "TREE":
    kwargs = {
        "PREFIX": PositionalParser(1),
        "FILES": PositionalParser("+")
    }
    return StandardArgTree.parse(ctx, tokens, 2, kwargs, [], breakstack)

  kwargs = {
      "FILES": PositionalParser("+"),
      "REGULAR_EXPRESSION": PositionalParser(1)
  }

  descriminator = get_nth_semantic_token(tokens, 1)
  if descriminator is not None and descriminator.spelling.upper() not in kwargs:
    return StandardArgTree.parse(ctx, tokens, 2, {}, [], breakstack)

  return StandardArgTree.parse(ctx, tokens, 1, kwargs, [], breakstack)


def populate_db(parse_db):
  parse_db["build_command"] = parse_build_command
  parse_db["cmake_parse_arguments"] = parse_cmake_parse_arguments
  parse_db["cmake_policy"] = parse_cmake_policy
  parse_db["create_test_sourcelist"] = parse_create_test_sourcelist
  parse_db["export_library_dependencies"] = parse_export_library_dependencies
  parse_db["fltk_wrap_ui"] = parse_fltk_wrap_ui
  parse_db["get_cmake_property"] = parse_get_cmake_property
  parse_db["get_filename_component"] = parse_get_filename_component
  parse_db["get_source_file_property"] = parse_get_source_file_property
  parse_db["get_target_property"] = parse_get_target_property
  parse_db["get_test_property"] = parse_get_test_property
  parse_db["include_external_msproject"] = parse_include_external_msproject
  parse_db["include_guard"] = parse_include_guard
  parse_db["include_regular_expression"] = parse_include_regular_expression
  parse_db["link_directories"] = parse_link_directories
  parse_db["link_libraries"] = parse_link_libraries
  parse_db["load_cache"] = parse_load_cache
  parse_db["load_command"] = parse_load_command
  parse_db["math"] = parse_math
  parse_db["option"] = parse_option
  parse_db["output_required_files"] = parse_output_required_files
  parse_db["qt_wrap_cpp"] = parse_qt_wrap_cpp
  parse_db["qt_wrap_ui"] = parse_qt_wrap_ui
  parse_db["remove_definitions"] = parse_remove_definitions
  parse_db["separate_arguments"] = parse_separate_arguments
  parse_db["set_source_files_properties"] = parse_set_source_files_properties
  parse_db["site_name"] = parse_site_name
  parse_db["source_group"] = parse_source_group



"""
Split help contents into the tool-specific short help and the common
configuration options.
"""

import argparse
import io

from cmakelang import configuration


def main():
  argparser = argparse.ArgumentParser(description=__doc__)
  argparser.add_argument("infile")
  argparser.add_argument("unique_outfile")
  argparser.add_argument("common_outfile", nargs="?", default="/dev/null")
  args = argparser.parse_args()

  matchlines = []
  for _, item in vars(configuration).items():
    if (isinstance(item, type) and
        issubclass(item, configuration.ConfigObject) and
        item is not configuration.ConfigObject):
      firstline = item.__doc__.strip().split("\n")[0]
      matchlines.append(firstline)

  unique_lines = []
  common_lines = []
  with io.open(args.infile, "r", encoding="utf-8") as infile:
    for line in infile:
      if line.strip().rstrip(":") in matchlines:
        common_lines.append(line)
        break
      unique_lines.append(line)

    for line in infile:
      common_lines.append(line)

  with io.open(args.unique_outfile, "w", encoding="utf-8") as outfile:
    outfile.write("".join(unique_lines))

  if args.common_outfile != "/dev/null":
    with io.open(args.common_outfile, "w", encoding="utf-8") as outfile:
      outfile.write("".join(common_lines))


if __name__ == "__main__":
  main()



# pylint: disable=bad-continuation
# pylint: disable=too-many-lines

FUNSPECS = {
  "add_test": {
    "kwargs": {
      "COMMAND": {
        "pargs": "+",
        "kwargs": {
          "ARGS": "*"
        }
      },
      "CONFIGURATIONS": "*",
      "NAME": "*",
      "WORKING_DIRECTORY": "*"
    }
  },
  "cmake_host_system_information": {
    "kwargs": {
      "QUERY": "*",
      "RESULT": "*"
    }
  },
  "cmake_minimum_required": {
    "flags": [
      "FATAL_ERROR"
    ],
    "kwargs": {
      "VERSION": "*"
    }
  },
  "cmake_minimum_required_version": {
    "kwargs": {
      "VERSION": "*"
    }
  },
  "configure_file": {
    "flags": [
      "@ONLY",
      "COPYONLY",
      "ESCAPE_QUOTES"
    ],
    "kwargs": {
      "NEWLINE_STYLE": "*"
    }
  },
  "define_property": {
    "flags": [
      "CACHED_VARIABLE",
      "DIRECTORY",
      "GLOBAL",
      "INHERITED",
      "SOURCE",
      "TARGET",
      "TEST",
      "VARIABLE"
    ],
    "kwargs": {
      "BRIEF_DOCS": "*",
      "FULL_DOCS": "*",
      "PROPERTY": "*"
    }
  },
  "enable_language": {
    "flags": [
      "OPTIONAL"
    ],
    "kwargs": {}
  },
  "execute_process": {
    "flags": [
      "ERROR_QUIET",
      "ERROR_STRIP_TRAILING_WHITESPACE",
      "OUTPUT_QUIET",
      "OUTPUT_STRIP_TRAILING_WHITESPACE"
    ],
    "kwargs": {
      "COMMAND": "*",
      "ERROR_FILE": "*",
      "ERROR_VARIABLE": "*",
      "INPUT_FILE": "*",
      "OUTPUT_FILE": "*",
      "OUTPUT_VARIABLE": "*",
      "RESULT_VARIABLE": "*",
      "TIMEOUT": "*",
      "WORKING_DIRECTORY": "*"
    }
  },
  "export": {
    "flags": [
      "APPEND",
      "EXPORT_LINK_INTERFACE_LIBRARIES"
    ],
    "kwargs": {
      "ANDROID_MK": 1,
      "EXPORT": 1,
      "FILE": 1,
      "NAMESPACE": 1,
      "PACKAGE": 1,
      "TARGETS": "+"
    }
  },
  "find_file": {
    "flags": [
      "CMAKE_FIND_ROOT_PATH_BOTH",
      "NO_CMAKE_ENVIRONMENT_PATH",
      "NO_CMAKE_FIND_ROOT_PATH",
      "NO_CMAKE_PATH",
      "NO_CMAKE_SYSTEM_PATH",
      "NO_DEFAULT_PATH",
      "NO_SYSTEM_ENVIRONMENT_PATH",
      "ONLY_CMAKE_FIND_ROOT_PATH"
    ],
    "kwargs": {
      "DOC": "*",
      "HINTS": "*",
      "NAMES": "*",
      "PATHS": "*",
      "PATH_SUFFIXES": "*"
    }
  },
  "find_library": {
    "flags": [
      "CMAKE_FIND_ROOT_PATH_BOTH",
      "NO_CMAKE_ENVIRONMENT_PATH",
      "NO_CMAKE_FIND_ROOT_PATH",
      "NO_CMAKE_PATH",
      "NO_CMKE_SYSTEM_PATH",
      "NO_DEFAULT_PATH",
      "NO_SYSTEM_ENVIRONMENT_PATH",
      "ONLY_CMAKE_FIND_ROOT_PATH"
    ],
    "kwargs": {
      "DOC": "*",
      "HINTS": "*",
      "NAMES": "*",
      "PATHS": "*",
      "PATH_SUFFIXES": "*"
    }
  },
  "find_package": {
    "pargs": "*",
    "flags": [
      "EXACT",
      "MODULE",
      "REQUIRED",
      "NO_POLICY_SCOPE",
      "QUIET"
    ],
    "kwargs": {
      "COMPONENTS": "*",
      "OPTIONAL_COMPONENTS": "*"
    }
  },
  "find_path": {
    "flags": [
      "CMAKE_FIND_ROOT_PATH_BOTH",
      "NO_CMAKE_ENVIRONMENT_PATH",
      "NO_CMAKE_FIND_ROOT_PATH",
      "NO_CMAKE_PATH",
      "NO_CMKE_SYSTEM_PATH",
      "NO_DEFAULT_PATH",
      "NO_SYSTEM_ENVIRONMENT_PATH",
      "ONLY_CMAKE_FIND_ROOT_PATH"
    ],
    "kwargs": {
      "DOC": "*",
      "HINTS": "*",
      "NAMES": "*",
      "PATHS": "*",
      "PATH_SUFFIXES": "*"
    }
  },
  "find_program": {
    "flags": [
      "CMAKE_FIND_ROOT_PATH_BOTH",
      "NO_CMAKE_ENVIRONMENT_PATH",
      "NO_CMAKE_FIND_ROOT_PATH",
      "NO_CMAKE_PATH",
      "NO_CMAKE_SYSTEM_PATH",
      "NO_DEFAULT_PATH",
      "NO_SYSTEM_ENVIRONMENT_PATH",
      "ONLY_CMAKE_FIND_ROOT_PATH"
    ],
    "kwargs": {
      "DOC": "*",
      "HINTS": "*",
      "NAMES": "*",
      "PATHS": "*",
      "PATH_SUFFIXES": "*"
    }
  },
  "get_directory_property": {
    "kwargs": {
      "DIRECTORY": "*"
    }
  },
  "get_property": {
    "flags": [
      "BRIEF_DOCS",
      "DEFINED",
      "FULL_DOCS",
      "GLOBAL",
      "SET",
      "VARIABLE"
    ],
    "kwargs": {
      "CACHE": "*",
      "DIRECTORY": "*",
      "PROPERTY": "*",
      "SOURCE": "*",
      "TARGET": "*",
      "TEST": "*"
    }
  },
  "include": {
    "flags": [
      "NO_POLICY_SCOPE",
      "OPTIONAL"
    ],
    "kwargs": {
      "RESULT_VARIABLE": "*"
    }
  },
  "include_directories": {
    "flags": [
      "AFTER",
      "BEFORE",
      "SYSTEM"
    ],
    "kwargs": {}
  },
  "mark_as_advanced": {
    "flags": [
      "CLEAR",
      "FORCE"
    ],
    "kwargs": {}
  },
  "message": {
    "kwargs": {
      "AUTHOR_WARNING": "*",
      "DEPRECATION": "*",
      "FATAL_ERROR": "*",
      "SEND_ERROR": "*",
      "STATUS": "*",
      "WARNING": "*"
    }
  },
  "project": {
    "kwargs": {
      "DESCRIPTION": 1,
      "HOMEPAGE_URL": 1,
      "LANGUAGES": "*",
      "VERSION": "*"
    }
  },
  "set_directory_properties": {
    "kwargs": {
      "PROPERTIES": "*"
    }
  },
  "set_property": {
    "flags": [
      "APPEND",
      "APPEND_STRING",
      "GLOBAL"
    ],
    "kwargs": {
      "CACHE": "*",
      "DIRECTORY": "*",
      "PROPERTY": "*",
      "SOURCE": "*",
      "TARGET": "*",
      "TEST": "*"
    }
  },
  "set_tests_properties": {
    "kwargs": {
      "PROPERTIES": "*"
    }
  },
  "string": {
    "flags": [
      "@ONLY",
      "ESCAPE_QUOTES",
      "REVERSE",
      "UTC"
    ],
    "kwargs": {
      "ALPHABET": "*",
      "ASCII": "*",
      "COMPARE": "*",
      "CONCAT": "*",
      "CONFIGURE": "*",
      "FIND": "*",
      "LENGTH": "*",
      "MAKE_C_IDENTIFIER": "*",
      "MD5": "*",
      "RANDOM": "*",
      "RANDOM_SEED": "*",
      "REGEX": "*",
      "REPLACE": "*",
      "SHA1": "*",
      "SHA256": "*",
      "SHA384": "*",
      "SHA512": "*",
      "STRIP": "*",
      "SUBSTRING": "*",
      "TIMESTAMP": "*",
      "TOLOWER": "*",
      "TOUPPER": "*"
    }
  },
  "try_compile": {
    "pargs": [2, "+"],
    "kwargs": {
      "CMAKE_FLAGS": "*",
      "COMPILE_DEFINITIONS": "*",
      "COPY_FILE": "*",
      "LINK_OPTIONS": "*",
      "LINK_LIBRARIES": "*",
      "OUTPUT_VARIABLE": "*",
      "RESULT_VAR": "*",
      "SOURCES": "*"
    }
  },
  "target_compile_options": {
    "kwargs": {
      "PRIVATE": "+",
      "PUBLIC": "+",
      "INTERFACE": "+"
    }
  },
  "target_include_directories": {
    "kwargs": {
      "PRIVATE": "+",
      "PUBLIC": "+",
      "INTERFACE": "+"
    }
  },
  "target_link_libraries": {
    "kwargs": {
      "PRIVATE": "+",
      "PUBLIC": "+",
      "INTERFACE": "+"
    }
  },
  "try_run": {
    "kwargs": {
      "ARGS": "*",
      "CMAKE_FLAGS": "*",
      "COMPILE_DEFINITIONS": "*",
      "COMPILE_OUTPUT_VARIABLE": "*",
      "OUTPUT_VARIABLE": "*",
      "RUN_OUTPUT_VARIABLE": "*"
    }
  }
}



"""
Get the version string by parsing a C++ header file. The version is expected
to be in the form of:

#define FOO_VERSION \
  { 0, 1, 0, "dev", 0 }
"""

from __future__ import unicode_literals

import argparse
import os
import io
import sys

OUTFILE_TPL = """
set({macro} "{semver}")
set({prefix}_API_VERSION "{api_version}")
set({prefix}_SO_VERSION "{so_version}")
"""


def main():
  argparser = argparse.ArgumentParser(description=__doc__)
  argparser.add_argument("-o", "--outfilepath", default="-")
  argparser.add_argument("macro")
  argparser.add_argument("filepath")
  args = argparser.parse_args()
  if args.outfilepath == "-":
    args.outfilepath = os.dup(sys.stdout.fileno())

  version = None
  with io.open(args.filepath, "r", encoding="utf-8") as infile:
    lineiter = iter(infile)
    for line in lineiter:
      if line.startswith("#define " + args.macro):
        version = next(lineiter)
        break

  if version is None:
    return 1
  # Strip whitespace and brackets:
  version = version.strip()[1:-1]
  # Split along commas
  verparts = [part.strip().strip('"') for part in version.split(",")]

  prefix = args.macro
  if not prefix.endswith("_VERSION"):
    return 1

  prefix = prefix[:-len("_VERSION")]
  outputs = {
      "macro": args.macro,
      "prefix": prefix,
      "so_version": ".".join(verparts[:2]),
      "api_version": ".".join(verparts[:3]),
  }

  semver = ".".join(verparts[:3])
  pre_release = "".join(verparts[3:])
  if pre_release:
    semver += "-" + pre_release
  outputs["semver"] = semver

  with io.open(args.outfilepath, "w", encoding="utf-8") as outfile:
    outfile.write(OUTFILE_TPL.format(**outputs))
  return 0


if __name__ == "__main__":
  sys.exit(main())
